/****************************************************************
Copyright (c) 1998 Sandro Sigala <ssigala@globalnet.it>.
All rights reserved.

Permission to use, copy, modify, and distribute this software
and its documentation for any purpose and without fee is hereby
granted, provided that the above copyright notice appear in all
copies and that both that the copyright notice and this
permission notice and warranty disclaimer appear in supporting
documentation, and that the name of the author not be used in
advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

The author disclaim all warranties with regard to this
software, including all implied warranties of merchantability
and fitness.  In no event shall the author be liable for any
special, indirect or consequential damages or any damages
whatsoever resulting from loss of use, data or profits, whether
in an action of contract, negligence or other tortious action,
arising out of or in connection with the use or performance of
this software.
****************************************************************/

#include "config.h"

#include <kapp.h>
#include <ktopwidget.h>
#include <kmsgbox.h>
#include <qpopmenu.h>
#include <qaccel.h>
#include <qlistbox.h>
#include <qpainter.h>

#include "mainwindow.h"
#include "mainwidget.h"
#include "spritewindow.h"
#include "mapwindow.h"
#include "newmapdialog.h"
#include "newtransporterdialog.h"

MainWindow::MainWindow(const char *name)
	: KTopLevelWidget(name)
{
	setCaption("KRepton Editor");

	view = new MainWidget(this);
	setView(view);

	file_popup = new QPopupMenu();
	file_popup->insertItem("&Save", this, SLOT(menu_save()));
	file_popup->insertSeparator();
	file_popup->insertItem("Strict &Check", this, SLOT(menu_strictCheck()), Key_F10);
	file_popup->insertSeparator();
	file_popup->insertItem("&Quit", qApp, SLOT(quit()), CTRL+Key_Q);

	options_popup = new QPopupMenu();
	strictID = options_popup->insertItem("&Strict Check On Save", this, SLOT(menu_strictChecking()));
	alignID = options_popup->insertItem("&Align Windows", this, SLOT(menu_alignWindows()));

	windows_popup = new QPopupMenu();
	windows_popup->insertItem("&Sprite Editor", this, SLOT(menu_spriteEditor()), Key_F2);
	windows_popup->insertItem("&Level Editor", this, SLOT(menu_levelEditor()), Key_F3);
	windows_popup->insertSeparator();
	windows_popup->insertItem("&Realign Windows", this, SLOT(menu_realign()));

	help_popup = new QPopupMenu();
	help_popup->insertItem("&Contents", this, SLOT(menu_help()), Key_F1);
	help_popup->insertSeparator();
	help_popup->insertItem("&About KRepton Editor...", this, SLOT(menu_about()));
//	help_popup->insertItem("About &KDE...", this, SLOT(menu_aboutKDE()));

	menu = new KMenuBar(this);
	menu->insertItem("&File", file_popup);
	menu->insertItem("&Options", options_popup);
	menu->insertItem("&Windows", windows_popup);
	menu->insertItem("&Help", help_popup);
	setMenu(menu);

	sprites = new QPixmap[num_sprites];
	sprite_window = new SpriteWindow(sprites);
	map_window = new MapWindow(sprites);

	connect(view->maps_list, SIGNAL(highlighted(int)),
		this, SLOT(selectedMap(int)));
	connect(view->password_edit, SIGNAL(textChanged(const char *)),
		this, SLOT(changedPassword(const char *)));
	connect(view->time_edit, SIGNAL(textChanged(const char *)),
		this, SLOT(changedTime(const char *)));
	connect(view->map_insert, SIGNAL(clicked()),
		this, SLOT(mapInsert()));
	connect(view->map_remove, SIGNAL(clicked()),
		this, SLOT(mapRemove()));
	connect(view->map_move_up, SIGNAL(clicked()),
		this, SLOT(mapMoveUp()));
	connect(view->map_move_down, SIGNAL(clicked()),
		this, SLOT(mapMoveDown()));
	connect(sprite_window, SIGNAL(changedSprite()),
		this, SLOT(changedSprite()));
	connect(view->transporter_insert, SIGNAL(clicked()),
		this, SLOT(transporterInsert()));
	connect(view->transporter_remove, SIGNAL(clicked()),
		this, SLOT(transporterRemove()));

	KConfig *config = kapp->getConfig();
	config->setGroup("Options");
	strict_checking = config->readBoolEntry("StrictChecking", true);
	options_popup->setItemChecked(strictID, strict_checking);
	auto_align = config->readBoolEntry("AutoAlign", true);
	options_popup->setItemChecked(alignID, auto_align);

//	showWindows();
}

MainWindow::~MainWindow()
{
}

void MainWindow::loadEpisode(const QString& path)
{
	episode_path = path;

	Map *map;
	while ((map = maps.first()) != 0)
		maps.remove();

	QString filename = path + "/info";
	QString name;
	int num_levels;

	QFile f(filename);
	if (!f.open(IO_ReadOnly))
		fatal("Cannot open %s", (const char *)filename);
	QTextStream t(&f);
	name = t.readLine();
	t >> num_levels;
	f.close();

	view->name_edit->setText(name);

	for (int i = 0; i < num_levels; ++i) {
		map = new Map();
		filename.sprintf("%s/map%d", (const char *)path, i+1);
		map->load(filename);
		maps.append(map);
	}

	updateMapsList();
	selectLevel(0);

	loadSprites(path);
}

void MainWindow::loadSprites(const QString& path)
{
	QPixmap allsprites;

	allsprites.load(path + "/sprites.bmp");

	for (int y = 0; y < 6; ++y)
		for (int x = 0; x < 8; ++x) {
			int i = y*8 + x; 
			sprites[i].resize(sprite_width, sprite_height);
			QPainter p;
			p.begin(&sprites[i]);
			p.drawPixmap(0, 0, allsprites,
				     x * sprite_width, y * sprite_height,
				     sprite_width, sprite_height);
			p.end();
		}
}

void MainWindow::updateMapsList()
{
	view->maps_list->setAutoUpdate(false);
	view->maps_list->clear();
	for (Map *m = maps.first(); m != 0; m = maps.next())
		view->maps_list->insertItem(m->password);
	view->maps_list->setAutoUpdate(true);
	view->maps_list->repaint();
}

void MainWindow::updateTransportersList(int level)
{
	Map *map = maps.at(level);

	view->transporters_list->setAutoUpdate(false);
	view->transporters_list->clear();
	for (Transporter *t = map->transporters.first(); t != 0;
	     t = map->transporters.next()) {
		QString s;
		s.sprintf("%d,%d to %d,%d",
			  t->orig_x+1, t->orig_y+1, t->dest_x+1, t->dest_y+1);
		view->transporters_list->insertItem(s);
	}
	if (map->transporters.count() > 0) {
		view->transporters_list->setCurrentItem(0);
		view->transporters_list->centerCurrentItem();
	}
	view->transporters_list->setAutoUpdate(true);
	view->transporters_list->repaint();
}

void MainWindow::selectLevel(int level)
{
	Map *map = maps.at(level);

	updateTransportersList(level);
	view->password_edit->setText(map->password);
	QString s;
	s.setNum(map->time_limit);
	view->time_edit->setText(s);
	map_window->setMap(map);
}

void MainWindow::showWindows()
{
	sprite_window->show();
	map_window->show();
	if (auto_align) alignWindows();
}

void MainWindow::alignWindows()
{
	QRect g = frameGeometry();
	sprite_window->move(x() + g.width(), y());
	QRect h = sprite_window->frameGeometry();
	map_window->move(x() + g.width(), sprite_window->y() + h.height());
}

void MainWindow::mapInsert()
{
	NewMapDialog d(this);

	for (;;) {
		if (!d.exec())
			return;
		if (d.mapPassword() == "")
			KMsgBox::message(this, "KRepton Editor",
					 "You must specify a password for the level!");
		else if (d.mapWidth() < 10)
			KMsgBox::message(this, "KRepton Editor",
					 "The minimum width is 10!");
		else if (d.mapHeight() < 10)
			KMsgBox::message(this, "KRepton Editor",
					 "The minimum height is 10!");
		else
			break;
	}

	Map *m = new Map();
	m->newMap(d.mapWidth(), d.mapHeight());
	m->password = d.mapPassword();
	maps.append(m);
	updateMapsList();
	view->maps_list->setCurrentItem(maps.count() - 1);
	view->maps_list->centerCurrentItem();
}

void MainWindow::mapRemove()
{
	int item = view->maps_list->currentItem();
	if (maps.count() == 1)
		KMsgBox::message(this, "KRepton Editor",
				 "You cannot remove the latest level; first insert another level, then remove this!");
	else {
		QString msg;
		msg.sprintf("Are you sure to remove the map #%d?", item+1);
		if (KMsgBox::yesNo(this, "KRepton Editor", msg) == 1) {
			Map *m = maps.at(item);
			maps.remove(item);
			delete m;
			updateMapsList();
			view->maps_list->setCurrentItem(0);
			view->maps_list->centerCurrentItem();
		}
	}
}

void MainWindow::mapMoveUp()
{
	int item = view->maps_list->currentItem();
	if (item == 0)
		KMsgBox::message(this, "KRepton Editor",
				 "You cannot move the first level one position up!");
	else {
		Map *m = maps.at(item);
		maps.remove(item);
		maps.insert(item-1, m);
		updateMapsList();
		view->maps_list->setCurrentItem(item-1);
		view->maps_list->centerCurrentItem();
	}
}

void MainWindow::mapMoveDown()
{
	int item = view->maps_list->currentItem();
	int last_item = maps.count() - 1;
	if (item == last_item)
		KMsgBox::message(this, "KRepton Editor",
				 "You cannot move the last level one position down!");
	else {
		Map *m = maps.at(item);
		maps.remove(item);
		maps.insert(item+1, m);
		updateMapsList();
		view->maps_list->setCurrentItem(item+1);
		view->maps_list->centerCurrentItem();
	}
}

void MainWindow::transporterInsert()
{
	Map *map = maps.at(view->maps_list->currentItem());
	QString swidth, sheight;

	swidth.setNum(map->width);
	sheight.setNum(map->height);

	NewTransporterDialog d(this);

	for (;;) {
		if (!d.exec())
			return;
		if (d.xorig() < 1)
			KMsgBox::message(this, "KRepton Editor",
					 "The minimum X origin is 1!");
		else if (d.yorig() < 1)
			KMsgBox::message(this, "KRepton Editor",
					 "The minimum Y origin is 1!");
		else if (d.xdest() < 1)
			KMsgBox::message(this, "KRepton Editor",
					 "The minimum X destination is 1!");
		else if (d.ydest() < 1)
			KMsgBox::message(this, "KRepton Editor",
					 "The minimum Y destination is 1!");
		else if (d.xorig() > map->width)
			KMsgBox::message(this, "KRepton Editor",
					 "The maximum X origin is "
					 + swidth + "!");
		else if (d.yorig() > map->height)
			KMsgBox::message(this, "KRepton Editor",
					 "The maximum Y origin is "
					 + sheight + "!");
		else if (d.xdest() > map->width)
			KMsgBox::message(this, "KRepton Editor",
					 "The maximum X destination is "
					 + swidth + "!");
		else if (d.ydest() > map->height)
			KMsgBox::message(this, "KRepton Editor",
					 "The maximum Y destination is "
					 + sheight + "!");
		else
			break;
	}

	Transporter *t = new Transporter(d.xorig()-1, d.yorig()-1,
					 d.xdest()-1, d.ydest()-1);
	map->transporters.append(t);
	updateTransportersList(view->maps_list->currentItem());
	view->transporters_list->setCurrentItem(map->transporters.count() - 1);
	view->transporters_list->centerCurrentItem();
}

void MainWindow::transporterRemove()
{
	Map *map = maps.at(view->maps_list->currentItem());

	if (map->transporters.count() < 1) {
		KMsgBox::message(this, "KRepton Editor",
				 "There are no transporters to remove!");
		return;
	}

	int item = view->transporters_list->currentItem();
	QString msg;
	msg.sprintf("Are you sure to remove the transporter #%d?", item+1);
	if (KMsgBox::yesNo(this, "KRepton Editor", msg) == 1) {
		Transporter *t = map->transporters.at(item);
		map->transporters.remove(item);
		delete t;
		updateTransportersList(view->maps_list->currentItem());
	}
}

void MainWindow::selectedMap(int level)
{
	selectLevel(level);
}

void MainWindow::changedPassword(const char *s)
{
	int level = view->maps_list->currentItem();
	if (level >= 0) {
		view->maps_list->changeItem(s, level);
		maps.at(level)->password = s;
	}
}

void MainWindow::changedTime(const char *s)
{
	int level = view->maps_list->currentItem();
	if (level >= 0)
		maps.at(level)->time_limit = QString(s).toInt();
}

void MainWindow::changedSprite()
{
	sprite_window->updateChilds();
	map_window->updateChilds();
}

bool MainWindow::strictCheckTransporter(Map *map, const QString& thismap, int x, int y)
{
	int found = 0;
	for (Transporter *t = map->transporters.first(); t != 0; t = map->transporters.next())
		if (t->orig_x == x && t->orig_y == y)
			++found;

	QString pos;
	pos.sprintf("[X:%d, Y:%d]", x+1, y+1);

	if (found == 0) {
		KMsgBox::message(this, "KRepton Editor",
				 "Cannot save: the map #" + thismap + " has a transporter at position " + pos + " with no specified destination!",
				 KMsgBox::EXCLAMATION);
		return false;
	}
	if (found > 1) {
		KMsgBox::message(this, "KRepton Editor",
				 "Cannot save: the map #" + thismap + " has a transporter at position " + pos + " with more that one specified destination!",
				 KMsgBox::EXCLAMATION);
		return false;
	}

	return true;
}

bool MainWindow::strictCheckTransporters(Map *map, const QString& thismap)
{
	for (Transporter *t = map->transporters.first(); t != 0; t = map->transporters.next()) {
		QString num;
		num.setNum(map->transporters.at() + 1);
		if (map->xy(t->orig_x, t->orig_y) != Sprite_Transport) {
			KMsgBox::message(this, "KRepton Editor",
					 "Cannot save: the transporter #" + num + " of the map #"
					 + thismap + " has an invalid origin (no transporter sprite was found)!",
					 KMsgBox::EXCLAMATION);
			return false;
		}
	}

	return true;
}

bool MainWindow::strictCheck()
{
	for (Map *map = maps.first(); map != 0; map = maps.next()) {
		QString thismap;
		thismap.setNum(maps.at() + 1);

		if (map->password == "") {
			KMsgBox::message(this, "KRepton Editor",
					 "Cannot save: the map #" + thismap + " has no password specified!",
					 KMsgBox::EXCLAMATION);
			return false;
 		}

 		if (map->time_limit <= 0) {
			KMsgBox::message(this, "KRepton Editor",
					 "Cannot save: the map #" + thismap + " has an invalid time limit!",
					 KMsgBox::EXCLAMATION);
			return false;
		}

		int num_reptons = 0, num_diamonds = 0;
		int num_blips = 0, num_cages = 0, num_bombs = 0;
		int num_strongboxes = 0, num_keys = 0;
		for (int y = 0; y < map->height; ++y)
			for (int x = 0; x < map->width; ++x)
				switch (map->xy(x, y)) {
				case Sprite_Repton:
					++num_reptons;
					break;
				case Sprite_Diamond:
					++num_diamonds;
					break;
				case Sprite_Blip:
					++num_blips;
					break;
				case Sprite_Cage:
					++num_cages;
					break;
				case Sprite_Bomb:
					++num_bombs;
					break;
				case Sprite_Strongbox:
					++num_strongboxes;
					break;
				case Sprite_Key:
					++num_keys;
					break;
				case Sprite_Transport:
					if (!strictCheckTransporter(map, thismap, x, y))
						return false;
					break;
				default:; // Avoid warnings
				}

 		if (num_reptons == 0) {
			KMsgBox::message(this, "KRepton Editor",
					 "Cannot save: the map #" + thismap + " has no Repton!",
					 KMsgBox::EXCLAMATION);
			return false;
		}

 		if (num_reptons > 1) {
			KMsgBox::message(this, "KRepton Editor",
					 "Cannot save: the map #" + thismap + " has more than one Repton!",
					 KMsgBox::EXCLAMATION);
			return false;
		}

 		if (num_bombs == 0) {
			KMsgBox::message(this, "KRepton Editor",
					 "Cannot save: the map #" + thismap + " has no end level object!",
					 KMsgBox::EXCLAMATION);
			return false;
		}

 		if (num_cages > num_blips) {
			KMsgBox::message(this, "KRepton Editor",
					 "Cannot save: the map #" + thismap + " has more cages than blips!",
					 KMsgBox::EXCLAMATION);
			return false;
		}

 		if (num_strongboxes > 0 && num_keys < 1) {
			KMsgBox::message(this, "KRepton Editor",
					 "Cannot save: the map #" + thismap + " has strongboxes but no keys!",
					 KMsgBox::EXCLAMATION);
			return false;
		}

		if (!strictCheckTransporters(map, thismap))
			return false;
	}

	// Check for duplicated passwords
	QList<QString> pwds;
	for (Map *map = maps.first(); map != 0; map = maps.next())
		pwds.append(new QString(map->password));
	for (Map *map = maps.first(); map != 0; map = maps.next()) {
		int no = 0;
		for (QString *s = pwds.first(); s != 0; s = pwds.next())
			if (*s == map->password) {
				++no;
				if (no > 1) {
					QString map1, map2;
					map1.setNum(maps.at() + 1);
					map2.setNum(pwds.at() + 1);
					KMsgBox::message(this, "KRepton Editor",
							 "Cannot save: the map #" + map2 + " has the same password as the map #" 
							 + map1 + "!",
							 KMsgBox::EXCLAMATION);
					return false;
				}
			}
	}

	return true;
}

void MainWindow::menu_save()
{
	// Performs consistency checks
	if (QString(view->name_edit->text()) == "") {
		KMsgBox::message(this, "KRepton Editor",
				 "Cannot save: the episode has no name!",
				 KMsgBox::EXCLAMATION);
		return;
	}

	if (strict_checking)
		if (!strictCheck())
			return;

	QString filename;
	filename = episode_path + "/info";
	QFile f(filename);
	if (!f.open(IO_WriteOnly))
		fatal("Cannot open %s for writing", (const char *)filename);
	QTextStream t(&f);
	t << view->name_edit->text() << '\n';
	t << maps.count() << '\n';
	f.close();

	removeMapFiles(episode_path);

	for (Map *map = maps.first(); map != 0; map = maps.next()) {
		filename.sprintf("%s/map%d", (const char *)episode_path,
				 maps.at() + 1);
		map->save(filename);
	}

	QPixmap allsprites(8 * 32, 6 * 32);
	QPainter p;
	p.begin(&allsprites);
	for (int y = 0; y < 6; ++y)
		for (int x = 0; x < 8; ++x)
			p.drawPixmap(x*32, y*32, sprites[y*8 + x]);
	p.end();

	allsprites.save(episode_path + "/sprites.bmp", "BMP");
}

void MainWindow::menu_strictCheck()
{
	if (strictCheck())
		KMsgBox::message(this, "KRepton Editor", "Strict check successfull!");
}

void MainWindow::menu_strictChecking()
{
	strict_checking = !strict_checking;
	options_popup->setItemChecked(strictID, strict_checking);
	KConfig *config = kapp->getConfig();
	config->setGroup("Options");
	config->writeEntry("StrictChecking", strict_checking);
}

void MainWindow::menu_alignWindows()
{
	auto_align = !auto_align;
	options_popup->setItemChecked(alignID, auto_align);
	KConfig *config = kapp->getConfig();
	config->setGroup("Options");
	config->writeEntry("AutoAlign", auto_align);
}

void MainWindow::menu_spriteEditor()
{
	sprite_window->show();
	if (auto_align) alignWindows();
}

void MainWindow::menu_levelEditor()
{
	map_window->show();
	if (auto_align) alignWindows();
}

void MainWindow::menu_realign()
{
	alignWindows();
}

void MainWindow::menu_help()
{
	kapp->invokeHTMLHelp("", "");
}

void MainWindow::menu_about()
{
	KMsgBox::message(this, "About KRepton Editor",
			 "KRepton Editor " VERSION "\n\n"
			 "by Sandro Sigala <ssigala@globalnet.it>\n\n"
			 "Feedbacks are welcome!\n",
			 KMsgBox::INFORMATION);
}

void MainWindow::menu_aboutKDE()
{
}

void MainWindow::moveEvent(QMoveEvent *e)
{
	KTopLevelWidget::moveEvent(e);
	if (auto_align) alignWindows();
}
